@@ -194,6 +194,22 @@
     "Display MESSAGE temporarily if non-nil while BODY is evaluated."
     `(progn ,@body)))
 
+;; `condition-case-unless-debug' is introduced with Emacs 24.
+(if (fboundp 'condition-case-unless-debug)
+    (defalias 'tramp-compat-condition-case-unless-debug
+      'condition-case-unless-debug)
+  (defmacro tramp-compat-condition-case-unless-debug
+    (var bodyform &rest handlers)
+  "Like `condition-case' except that it does not catch anything when debugging."
+    (declare (debug condition-case) (indent 2))
+    (let ((bodysym (make-symbol "body")))
+      `(let ((,bodysym (lambda () ,bodyform)))
+	 (if debug-on-error
+	     (funcall ,bodysym)
+	   (condition-case ,var
+	       (funcall ,bodysym)
+	     ,@handlers))))))
+
 ;; `font-lock-add-keywords' does not exist in XEmacs.
 (defun tramp-compat-font-lock-add-keywords (mode keywords &optional how)
   "Add highlighting KEYWORDS for MODE."
@@ -308,43 +324,49 @@ Not actually used.  Use `(format \"%o\" i)' instead?"
 ;; `copy-directory' is a new function in Emacs 23.2.  Implementation
 ;; is taken from there.
 (defun tramp-compat-copy-directory
-  (directory newname &optional keep-time parents)
+  (directory newname &optional keep-time parents copy-contents)
   "Make a copy of DIRECTORY (compat function)."
-  (if (fboundp 'copy-directory)
-      (tramp-compat-funcall 'copy-directory directory newname keep-time parents)
+  (condition-case nil
+      (tramp-compat-funcall
+       'copy-directory directory newname keep-time parents copy-contents)
 
-    ;; If `default-directory' is a remote directory, make sure we find
-    ;; its `copy-directory' handler.
-    (let ((handler (or (find-file-name-handler directory 'copy-directory)
-		       (find-file-name-handler newname 'copy-directory))))
-      (if handler
-	  (funcall handler 'copy-directory directory newname keep-time parents)
+    ;; `copy-directory' is either not implemented, or it does not
+    ;; support the the COPY-CONTENTS flag.  For the time being, we
+    ;; ignore COPY-CONTENTS as well.
 
-	;; Compute target name.
-	(setq directory (directory-file-name (expand-file-name directory))
-	      newname   (directory-file-name (expand-file-name newname)))
-	(if (and (file-directory-p newname)
-		 (not (string-equal (file-name-nondirectory directory)
-				    (file-name-nondirectory newname))))
-	    (setq newname
-		  (expand-file-name
-		   (file-name-nondirectory directory) newname)))
-	(if (not (file-directory-p newname)) (make-directory newname parents))
+    (error
+     ;; If `default-directory' is a remote directory, make sure we
+     ;; find its `copy-directory' handler.
+     (let ((handler (or (find-file-name-handler directory 'copy-directory)
+			(find-file-name-handler newname 'copy-directory))))
+       (if handler
+	   (funcall handler 'copy-directory directory newname keep-time parents)
 
-	;; Copy recursively.
-	(mapc
-	 (lambda (file)
-	   (if (file-directory-p file)
-	       (tramp-compat-copy-directory file newname keep-time parents)
-	     (copy-file file newname t keep-time)))
-	 ;; We do not want to delete "." and "..".
-	 (directory-files
-	  directory 'full "^\\([^.]\\|\\.\\([^.]\\|\\..\\)\\).*"))
+	 ;; Compute target name.
+	 (setq directory (directory-file-name (expand-file-name directory))
+	       newname   (directory-file-name (expand-file-name newname)))
+	 (if (and (file-directory-p newname)
+		  (not (string-equal (file-name-nondirectory directory)
+				     (file-name-nondirectory newname))))
+	     (setq newname
+		   (expand-file-name
+		    (file-name-nondirectory directory) newname)))
+	 (if (not (file-directory-p newname)) (make-directory newname parents))
 
-	;; Set directory attributes.
-	(set-file-modes newname (file-modes directory))
-	(if keep-time
-	    (set-file-times newname (nth 5 (file-attributes directory))))))))
+	 ;; Copy recursively.
+	 (mapc
+	  (lambda (file)
+	    (if (file-directory-p file)
+		(tramp-compat-copy-directory file newname keep-time parents)
+	      (copy-file file newname t keep-time)))
+	  ;; We do not want to delete "." and "..".
+	  (directory-files
+	   directory 'full "^\\([^.]\\|\\.\\([^.]\\|\\..\\)\\).*"))
+
+	 ;; Set directory attributes.
+	 (set-file-modes newname (file-modes directory))
+	 (if keep-time
+	     (set-file-times newname (nth 5 (file-attributes directory)))))))))
 
 ;; TRASH has been introduced with Emacs 24.1.
 (defun tramp-compat-delete-file (filename &optional trash)
