@@ -1,6 +1,6 @@
 ;;; tramp-efs.el --- Make EFS a foreign method in Tramp
 
-;; Copyright (C) 2003-2011 Free Software Foundation, Inc.
+;; Copyright (C) 2003-2012 Free Software Foundation, Inc.
 
 ;; Author: Kai Großjohann <kai.grossjohann@gmx.net>
 ;; Keywords: comm, processes
@@ -100,7 +100,7 @@ To be set in `tramp-efs-file-name-handler'.")
 	  "\\(" tramp-host-with-port-regexp "\\)"
 	  tramp-postfix-host-regexp))
 
-;; Still need to do efs-path-root-short-circuit-regexp.
+;; Still need to do `efs-path-root-short-circuit-regexp'.
 
 ;; Disable EFS from file-name-handler-alist.
 
@@ -110,10 +110,9 @@ To be set in `tramp-efs-file-name-handler'.")
 (defun tramp-disable-efs ()
   "Turn EFS off.
 This is useful for unified remoting.  See
-`tramp-file-name-structure-unified' and
-`tramp-file-name-structure-separate' for details.  Requests
-suitable for EFS will be forwarded to EFS.  Also see the
-variables `tramp-efs-method', `tramp-default-method', and
+`tramp-file-name-structure' for details.  Requests suitable for
+EFS will be forwarded to EFS.  Also see the variables
+`tramp-efs-method', `tramp-default-method', and
 `tramp-default-method-alist'.
 
 This function is not needed in Emacsen which include Tramp, but is
@@ -212,8 +211,8 @@ pass to the OPERATION."
 (defsubst tramp-efs-file-name-p (filename)
   "Check if it's a filename that should be forwarded to EFS."
   (when (string-match (nth 0 tramp-file-name-structure) filename)
-    (let ((v (tramp-dissect-file-name filename)))
-      (string= (tramp-file-name-method v) tramp-efs-method))))
+    (string= (tramp-file-name-method (tramp-dissect-file-name filename))
+	     tramp-efs-method)))
 
 ;;;###tramp-autoload
 (when (featurep 'xemacs)
